__version__ = "0.5.0"
__long_desc__ = """

# Swarmauri Experimental SDK

## Experimental
- **In-Development Components**: Early-stage features and components that push the boundaries of the Swarmauri framework, offering innovative solutions that are still in testing phases.

Visit us at: https://swarmauri.com
Follow us at: https://github.com/swarmauri 

"""